import axios, {
  AxiosInstance,
  AxiosRequestConfig,
  AxiosResponse,
  Canceler,
  CancelTokenSource,
  InternalAxiosRequestConfig,
} from "axios";
import qs from "qs";
import { AESCipher } from "./aesCipher";
import type {
  RequsetsServiceConfig,
  InternalRequestConfig,
  RequestConfig,
  CryptoConfig,
  // ContentType,
} from "./index.d";

let repeatRequests: Record<string, Canceler[]> = {};

function generateRequestKey(config: InternalRequestConfig) {
  const { method, url, params, data } = config;
  const split = "==";
  const array = [`url:`, url, `${split}method:`, method];
  params && array.push(`${split}params:`, qs.stringify(params));
  data &&
    array.push(
      `${split}data:`,
      typeof data === "object" ? JSON.stringify(data) : data
    );
  return array.join("");
}

function addPendingRequest(config: InternalRequestConfig) {
  const requestKey = generateRequestKey(config);
  config.cancelToken = new axios.CancelToken((cancel) => {
    if (!repeatRequests[requestKey]) {
      repeatRequests[requestKey] = [];
    }
    repeatRequests[requestKey].push(cancel);
  });
  return config;
}

function removePendingRequest(config: InternalRequestConfig) {
  const requestKey = generateRequestKey(config);
  const needCancel = repeatRequests[requestKey]?.length > 1;
  if (needCancel) {
    repeatRequests[requestKey].forEach((cancel) => {
      cancel(requestKey);
    });
  }
  needCancel ? (repeatRequests[requestKey] = []) : (repeatRequests = {});
}

function createRequestInstance(
  requestsServiceConfig: RequsetsServiceConfig
): Record<string, AxiosInstance> {
  const { baseURLs } = requestsServiceConfig;
  const instances: Record<string, AxiosInstance> = {};

  for (const key in baseURLs) {
    if (baseURLs.hasOwnProperty(key)) {
      const config = baseURLs[key];
      const {
        timeout,
        baseURL,
        cryptoConfig = {},
        // contentType = ContentType.JSON,
      } = config;
      const { encryptUrl, encryptParams, decryptRes, encryptKey } =
        cryptoConfig;
      let aesCipher = null;
      if ((encryptUrl || encryptParams) && encryptKey) {
        aesCipher = new AESCipher(encryptKey);
      }
      const request = axios.create({
        baseURL,
        timeout: timeout || 10000, // 设置请求超时时间
      });

      // 给请求添加拦截器
      request.interceptors.request.use((config: InternalRequestConfig) => {
        addPendingRequest(config);

        // 对请求参数进行加密
        if (aesCipher && encryptParams && config.encryptParams === false) {
          const dataStr =
            typeof config.data === "string"
              ? config.data
              : qs.stringify(config.data);
          config.data = aesCipher.encrypt(dataStr);
        }

        // 对URL进行加密
        if (aesCipher && encryptUrl) {
          config.url = aesCipher.encrypt(qs.stringify(config.url));
        }
        return config;
      });

      request.interceptors.response.use(
        (response: AxiosResponse) => {
          removePendingRequest(response.config);
          if (decryptRes) {
            return Promise.resolve(aesCipher?.decrypt(response.data));
          }
          return Promise.resolve(response.data);
        },
        (error: any) => {
          return Promise.reject(error);
        }
      );

      instances[key] = request;
    }
  }
  return instances;
}

export default createRequestInstance;
